//----------------------------------------------------------------------
// File:  c:\cbproj\AudioIO\aio_libs.h  ("Audio-I/O to Winamp")
// Author: Wolfgang Buescher (DL4YHF)
// Date:   2011-07-17
// Purpose: Common header for various implementations of an "Audio-I/O"-DLL,
//          which connects applications producing and consuming uncompressed
//          audio streams in real time .
//              (AIO_ReadInputSamplePoints,  AIO_WriteOutputSamplePoints) .
//          Shared memory objects are used for the data exchange.
//          The structures INSIDE the shared memory are defined here.
//      Note: The Audio-I/O-HOST doesn't need to know anything about
//            these structures; they are only used INSIDE the implementation
//            of certain audio-I/O-DLLs, for example in_AudioIO.dll .
//            The interface between the DLL and the DLL-HOST
//            is defined in  \cbproj\AudioIO\AudioIO.h (NOT in aio_libs.h) !
//
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
// Copyright, License, and Disclaimer (for the AudioIO-lib, excluding winamp plugin)
//      Copyright (c) 2008...2065, Wolfgang Bscher (DL4YHF)
//  Redistribution and use in source and binary forms, with or without
//  modification, are permitted provided that the following conditions are met:
//   1. Redistributions of source code must retain the above copyright notice,
//      this list of conditions and the following disclaimer.
//   2. Redistributions in binary form must reproduce the above copyright notice,
//      this list of conditions and the following disclaimer in the
//      documentation and/or other materials provided with the distribution.
//   3. Neither the name of the author nor the name of other contributors
//      may be used to endorse or promote products derived from this software
//      without specific prior written permission.
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
//  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
//   OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
//   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
//   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
// HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
// STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
// EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
//
// Revision history:
//  V0.02, 2011-10-31:  Created the header C:\cbproj\AudioIO\aio_libs.h ,
//                      which only contains 'DLL-internal' stuff
//         required for interaction between "aio2winamp.c" ("in_AudioIO.dll")
//         and the optional 'ASIO dummy' (ASIO_driver_in_c.c) in the same DLL .
//---------------------------------------------------------------------------

#define L_ALLOW_TESTING 1  /* 1=yes, 0=no */

#include <windows.h>   // contains stuff like "BOOL", etc
#include <math.h>
#include <string.h>
#include <stdio.h>   // no "standard I/O" but sprintf (a STRING function)
#pragma hdrstop      // no precompiled headers after this

#ifndef  _SWITCHES_H_
#  include "switches.h" // project-specific compilation switches,
                        // located in a project specific directory, like
                        // \cbproj\AudioIO\in_AudioIO\SWITCHES.H
                        // (the actual path is defined in the PROJECT OPTIONS)
#endif

#ifndef SWI_ASIO_SUPPORTED
# error "Missing definition of SWI_ASIO_SUPPORTED in SWITCHES.H !"
#endif
#if( SWI_ASIO_SUPPORTED ) // defined in SWITCHES.H (project-specific options)
# include "ASIO\Steinberg\asiosys.h" // must be copied from Steinberg's ASIO SDK !
# include "ASIO\Steinberg\asio.h"    // must be copied from Steinberg's ASIO SDK !
# include "ASIO\ASIO_dummy.h"        // once DL4YHF's "ASIO driver in plain C"
# include "ASIO\ASIO_Register.h"     // register / unregister our ASIO driver
#endif // def'd SWI_ASIO_SUPPORTED ?

//---------------------------------------------------------------------------
#include <shlobj.h>

#ifndef _CHUNK_INFO_H_
# include "ChunkInfo.h" // T_ChunkInfo in c:\cbproj\SoundUtl\ChunkInfo.h
#endif
#include "AudioIO.h" // Prototypes, data types and error codes for "Audio I/O"


//---------------------------------------------------------------------------

#define AUDIO_IO_MAX_CHANNELS 4


typedef struct tAudioReaderInfo // data for each AUDIO READER instance
{
  DWORD dwTailIndex; // incremented when READING data from the circular buffer
  DWORD dwFlags;     // bit combination of any of the following:
  #define AUDIO_READER_OPENED 0x00000001
  char  sz40ReaderID[44]; // short but hopefully unique name of this "reader",
                      // set through AIO_OpenAudioInput(pszAudioReaderID),
                      // possibly displayed on the DLL's configuration panel .
  int iDllInstanceIndex;  // to resolve ID-ambiguity in AIO_OpenAudioInput()
  int nChannelsPerSample;
  void  *pvUserDataForCallback;        // address of optional user data passed to the callback
  tpCallbackFromDLL pCallbackFunction; // address of an optional callback function
} T_AudioReader;
#define AUDIO_MAX_READERS 17

//---------------------------------------------------------------------------
// Structures for a shared memory block, for ONE writer and MULTIPLE readers,
//    only used inside this DLL .
// This WAS once a shared memory SECTION in the DLL;
//    declared with some strange, ugly, obfuscated stuff which only worked
//    under MinGW but not Borland ( __attribute__((section("shared"),shared)) ).
//    Unfortunately, the entire shared memory section was contained
//    in the DLL FILE, uncompressed, inflating its size to several megabytes.
// To reduce the size of the DLL, the shared memory section ("SharedData")
//    was later replaced with a shared memory object using OS functions like
//    'CreateFileMapping' / 'MapViewOfFile' / 'UnmapViewOfFile' / 'CloseHandle' .
//  For details, try to find the following articles at msdn :
//  - "Using Shared Memory in a Dynamic-Link Library"  :
//     http://msdn.microsoft.com/en-us/library/ms686958%28v=vs.85%29.aspx
//  - "Creating Named Shared Memory" :
//     http://msdn.microsoft.com/en-us/library/aa366551%28v=vs.85%29.aspx
//---------------------------------------------------------------------------

typedef struct tAudioSourceInfo  // "audio source info": part of T_SharedMemoryData ...
{ // Used in a shared memory array, saved in the config, and (usually) valid
  // BEFORE the 'writer' has opened/created a stream !
  char   sz40AudioStreamID[44];  // unique identifier (string) identifying THIS stream.
                                 // If sz40AudioStreamID[0] == '\0', this entry has NEVER been used yet.
  char   sz80Description[84];    // descriptive text, or simply the name of the application which 'generates' this stream
  double dblSampleRate;          // samples per second; PRECISE ("calibrated" if possible)
    // ex: int    iBitsPerSample;  // 16 bits per sample in most cases
    // (the structure of the 'shared buffer' is ALWAYS 32-bit floating point
    //  since 2011-07-01, to accomodate MULTIPLE readers)
  int    nChannelsPerSamplePoint;   // 1=mono, 2=stereo, 4=quadrophonic, etc
  double dblSampleValueRange;    // available even if no T_ChunkInfo is passed through the buffer .
         //  dblSampleValueRange = 1.0 means "samples are in the range -1.0 to +1.0"

  float  fltTestGeneratorFrequency; // frequency of the internal "test signal generator" in Hz
  DWORD  dwWriterFlags;   // just a **copy** of T_AudioBuffer.dwWriterFlags,
         // to make this member accessable through g_pSharedMemoryData->source[]
         // Contains bitflags like AUDIO_WRITER_OPENED, which allows ANY DLL
         // to find out quickly which of the audio streams are currently available,
         // even if their T_AudioBuffer (in another memory mapped file)
         // is not opened in 'this' instance of the DLL .

  // Info set by Audio Player (i.e. Winamp), may be important for the external source,
  // so we place it in the shared structure :
  int    iAudioPlayerInputPaused;   // Only for the AUDIO PLAYER (i.e. Winamp) !
  int    iAudioPlayerDisableOutput; // (can be set by external app, but affects Winamp)
  int    iAudioPlayerVolume;       // 0..255 (set by Winamp, read by others ... )
  int    iAudioPlayerBalance;      // -127..+127
  int    iAudioPlayerInitialized;  // TRUE: "winamp is there", FALSE: "winamp is not there"
  int    iAudioPlayerMode;    //  AIO_AUDIO_PLAYER_MODE_TONE, .._AUDI, ...  or ?
                              //  (set from a part of the "URL")


  // Diagnostics / parameters displayed on the control panel :
  float  fltAudioPeakValue[AUDIO_IO_MAX_CHANNELS]; // normalized, 0...1
  int    iCalculatedBufferUsage_0to255;
  int    iDisplayedBufferUsage_0to255;


  // Only for DISPLAY PURPOSES IN THE GUI, T_AudioBuffer.reader[]
  //      is duplicated in g_pSharedMemoryData->source[].reader[] :
  T_AudioReader reader[AUDIO_MAX_READERS];


} T_AudioSourceInfo;



typedef struct tMyAudioBuffer
{ // Structure of the "shared audio buffer".  Resides in a SHARED MEMORY object,
  //           common to ALL APPLICATIONS ("processes") which use this DLL,
  //           but also used internally by the ASIO-dummy (ASIO_dummy.c) .
  //   Some members which you may expect to find here (like sample rate,
  //   stream name, stream description, number of channels per sample point, etc)
  //   are only in T_SharedMemoryData.T_AudioSourceInfo source[] ,
  //                       i.e. in g_pSharedMemoryData->source[] .

  DWORD dwWriterFlags;   // bit combination of any of the following:
  #define AUDIO_WRITER_OPENED 0x00000002
  int    iWriterDllInstanceIndex;

  int    iLastErrorCode; // hopefully 0 = no error
  int    iResetRequest;  // flag to 'reset' the buffer somewhere,
         // most likely because the STRUCTURE (#channels, etc) has been modified


  // Head index for the circular buffer. Modified ONLY by the writer !
  DWORD dwHeadIndex; // incremented when WRITING data into the circular buffer (only one, 'cos there's only ONE writer)
                     // dwHeadIndex is incremented by <nChannelsPerSamplePoint>
                     //   for EACH NEW 'SAMPLE POINT' entered in the buffer !
       // Samples are entered as FLOATING POINT sample-groups,
       // and usually also read as FLOATING POINT sample-groups .
  int nChannelsPerSamplePoint; // copied from T_AudioSourceInfo for convenience
  int nSamplePointsMax;      // maximum number of SAMPLE POINTS which can be buffered..
  int nSamplePointsBuffered; // current number of sample points actually buffered..
                             // ..with source_info.nChannelsPerSamplePoint each
       // ex:  DWORD dwTailIndex; // incremented when READING data from the circular buffer
       // Since 2011-07-02, EACH 'READER' has it's own tail-index :
  T_AudioReader reader[AUDIO_MAX_READERS];
       // Note: for DISPLAY PURPOSES IN THE GUI, T_AudioBuffer.reader[]
       // is duplicated in the 'global' T_SharedMemoryData.T_AudioSourceInfo.reader[] .

  #define AUDIO_BUFFER_SIZE (256*1024) // max buffer capacity in 'single floats' (!)
    // MUST BE A POWER OF TWO (to simplify the buffer-wraps with a bitwise AND)
    //  Estimate: four channels in each sample point;
    //      buffer large enough for ONE SECOND of audio @ 192 kSamples/second;
    //      nearest power of two would be ~~ 256 * 1024 samples.
  float fltBuffer[AUDIO_BUFFER_SIZE]; // 2011-06-21: Buffer now consists of single-precision FLOATs !
  //  Remember, we cannot simply "malloc" a buffer dynamically because
  //  that buffer wouldn't be shared for all applications using the DLL !

  // Array of 'TChunkInfo' entries, added 2011-07-03 .
  // T_ChunkInfo is defined in c:\cbproj\SoundUtl\ChunkInfo.h .
  // Contains a precise timestamp, "calibrated" momentary sample rate,
  //           GPS data (when available),
  //           and other stuff required for experiments with TOA location .
  // How many 'chunk info blocks' are required ?
  //  Estimate: with AUDIO_BUFFER_SIZE == 256 * 1024,
  //            and the 'writer' emitting a chunk-info for each 1024 samples,
  //            256 chunk-infos stored in the buffer should be ok.
  //            sizeof(T_ChunkInfo) is approx. 256 byte, sum = 256*256 = 64 kByte.
  #define AUDIO_SAMPLES_PER_INFO_BLOCK 1024
  #define AUDIO_MAX_CHUNK_INFO_BLOCKS  (AUDIO_BUFFER_SIZE/AUDIO_SAMPLES_PER_INFO_BLOCK) /* 256 ? */
  T_ChunkInfo ChunkInfoArray[ AUDIO_MAX_CHUNK_INFO_BLOCKS ];
  int iLatestChunkInfoArrayIndex; // index of the LATEST entry in ChunkInfoArray[]; used by AudioBuf_GetLatestChunkInfo()
  // To retrieve the 'best possible' chunk-info for a given audio-sample quickly
  // ( based on the audio sample's index into fltBuffer[] ),
  //  T_AudioBuffer.dwHeadIndex divided by AUDIO_SAMPLES_PER_INFO_BLOCK
  //  will give an index into T_AudioBuffer.ChunkInfoLUT[] ("LUT"=Lookup-Table).
  // Details about this in  AudioBuf_GetChunkInfoIndexForSampleIndex() :
  struct // "chunk-info lookup table" :
   { int iAudioBufferIndex; // array index into fltBuffer[]
     int iChunkArrayIndex;  // REAL ("valid") array index into ChunkInfoArray[]
   } ChunkInfoLUT[AUDIO_MAX_CHUNK_INFO_BLOCKS];
} T_AudioBuffer;

typedef struct tControlPanelInfo
{
  int  iX1, iY1; // window position for the control panel
  int  iDisplayMode; // one of the following values:
#     define DISPLAY_MODE_SHOW_CONNECTIONS  0
#     define DISPLAY_MODE_SHOW_CHUNK_INFO   1
#     define DISPLAY_MODE_SHOW_WAVEFORM     2
#     define DISPLAY_MODE_SHOW_SPECTRUM     3
#     define DISPLAY_MODE_SHOW_EVT_HISTORY  4

  int   iMouseMode;  // what to do with a mouseclick.. one of THESE:
#     define MOUSE_MODE_OFF             0  // do nothing
#     define MOUSE_MODE_SEL_ASIO_STREAM 1  // select this stream for ASIO-I/O

} T_ControlPanelInfo;


typedef struct tEventHistoryEntry
{
  char sz255Text[256];
  SYSTEMTIME system_time; // date+time of entry, returned by Win32 : GetSystemTime()
} T_EventHistoryEntry;

typedef struct tSharedMemoryData
{
  int    iDllInstanceCounter;
  int    iGuiUpdateCounter;   // flag to update the GUI (display),
                              // possibly polled by MORE THAN ONE DLL-INSTANCE.

  #define AUDIO_MAX_SOURCES 8
  T_AudioSourceInfo source[AUDIO_MAX_SOURCES]; // source names, sampling rates, and other 'persistent' info

  int  iDayLastUsed[AUDIO_MAX_SOURCES];  // (unix-)day of last usage;
   // needed when running out of table entries to kick out the 'oldest' entry

  // ex: T_ControlPanelInfo cp_info;  // settings for the audio-I/O-DLL's "control panel" (GUI)
  // Since EACH INSTANCE may now show its own control panel (with different settings),
  //  the T_ControlPanelInfo data were removed from the 'shared memory block' in 2011-09 !

  // Winamp-specific stuff (as little as possible, here in T_SharedMemoryData):
  char sz255WinampLastEvent[256];   // can be displayed in ALL control panels
  int  iWinampCallFlags;
    #define WA_CALLED_GET_IN_MOD    0x01
    #define WA_CALLED_IS_OUR_FILE   0x02
    #define WA_CALLED_GET_FILE_INFO 0x04
    #define WA_CALLED_PLAY          0x08
    #define WA_CALLED_STOP          0x10
    #define WA_CALLED_PLAY_ERR      0x20
    #define WA_CALLED_PAUSE         0x40
    #define WA_CALLED_QUIT          0x80

  #define NUM_EVENT_HISTORY_ENTRIES 64 // max. number of history entries must be a power of two
  T_EventHistoryEntry EventHistory[NUM_EVENT_HISTORY_ENTRIES]; /* ca. 16 kByte */
  int  iEventHistoryHeadIndex ;  // Runs from 0 ... (NUM_EVENT_HISTORY_ENTRIES-1) .
  int  iEventHistoryNumEntries;  // Number of USED(!) entries in the event history.

} T_SharedMemoryData;

// Prototypes (not implementation-specific) :
__declspec( dllexport ) INT32 AIO_ShowControlPanel(
       HWND   hwndParent,      HWND   *phwndControlPanel,
       INT32  iDisplayOptions,
       void  *pvUserDataForCallback, tpCallbackFromDLL pCallbackFunction);
__declspec( dllexport ) INT32 AIO_OpenAudioInput( char *pszAudioStreamID,
       char *pszAudioReaderID,
       double  dblWantedSampleRate, double *pdblOutSampleRate,
       INT32 nChannelsPerSample, INT32 iTimeout_ms, INT32 iOpenOptions,
       T_ChunkInfo *pOutChunkInfo );
__declspec( dllexport )  void AIO_CloseAudioInput( INT32 iAudioIoHandle );



//---------------------------------------------------------------------------
void AppendToEventHistory( char *pszMessage );

//---------------------------------------------------------------------------
BOOL GetAudioBufferAndSourceInfoForWinamp(
               T_AudioBuffer **ppBuffer,       // [out] audio buffer to be used by Winamp
               T_AudioSourceInfo **ppSrcInfo); // [out] source-info to be used by Winamp

//---------------------------------------------------------------------------
T_AudioSourceInfo *GetAudioSourceInfoByStreamID( // may return NULL if "not found" !
        char *pszAudioStreamID , // [in] 'Stream ID' for the Audio-I/O-DLL, identifies the AUDIO SOURCE.
                                 //      use pszAudioStreamID=NULL to find the first UNUSED entry.
        int  *piAudioSource);    // [out,optional] INDEX of the audio source; 0..AUDIO_MAX_SOURCES-1

//---------------------------------------------------------------------------
long GetCurrentUnixDayNumber( void );
  // Used in this DLL to decide which of the entries in T_SharedMemoryData.source[]
  // has not been used for the longest time, and should be used when running
  // out of table entries .

//---------------------------------------------------------------------------
T_AudioBuffer * GetAudioBufferByStreamID(
        char * pszAudioStreamID, // [in,optional] stream ID' for the Audio-I/O-DLL, identifies the AUDIO SOURCE
                                 //      may be NULL to "use the first existing stream"
        int   iCreateOptions,    // [in] AUDIO_DONT_CREATE_BUFFER, AUDIO_CREATE_BUFFER_IF_NOT_EXISTING, ..
#       define AUDIO_DONT_CREATE_BUFFER            0x0000
#       define AUDIO_CREATE_BUFFER_IF_NOT_EXISTING 0x0001
        T_AudioSourceInfo **ppAudioSourceInfo, // [out,optional] address of the 'audio source info' for this stream
        int  *piAudioSource); // [out,optional] 0..AUDIO_MAX_SOURCES-1, will be part of the 'handle' later
  // Caution ! The result may be NULL if the specified audio source
  //           is unknown,  and couldn't be created (optionally) .


//---------------------------------------------------------------------------
T_AudioBuffer * GetAudioBufferByHandle( // may return NULL !
                          int iAudioIoHandle,  // [in] handle returned by .._OpenAudioOutput() / _OpenAudioInput()
                          int *piSourceIndex,  // [out,optional] index into T_SharedMemoryData.source[]
                          int *piReaderIndex,  // [out,optional] index into T_AudioBuffer.reader[]
       T_AudioSourceInfo **ppAudioSourceInfo); // [out,optional] address of the audio-source info
  // iAudioIoHandle identifies both, the audio WRITER (source) and the audio READER.
  // It doesn't have anything to do with a windows-handle, or a file-handle !
  // Here (inside aio2winamp),  the "audio source index" is in the upper bits,
  //    and the index into T_AudioBuffer. T_AudioReader reader[]
  //    is in the lower 8 bits (even though we don't need such a lot) .
  // GetAudioBufferByHandle() may return a NULL pointer if the handle
  //    is obviously invalid, or if the buffer doesn't exist anymore !

//---------------------------------------------------------------------------
INT32 GetNumberOfFreeSamplePointsInOutputBuffer(
                     int iAudioSource,    // [in]
               int *piNumActiveReaders ); // [out,optional]
  // Returns the number of 'unused' entries in T_AudioBuffer.fltBuffer[]
  // for the SLOWEST (but active) reader of this buffer.
  // If no reader is active, the return value indicates an 'empty' buffer.


// EOF <aio2winamp.cpp>




